from dataclasses import dataclass, field
from typing import TYPE_CHECKING, List, Union

import numpy as np
import pandas as pd
from biofit.processing import SelectedColumnTypes, sync_backup_config
from biofit.utils import logging
from sklearn.decomposition import PCA

from ..feature_extraction import FeatureExtractor, FeatureExtractorConfig

if TYPE_CHECKING:
    import polars as pl

logger = logging.get_logger(__name__)


@dataclass
class PCAFeatureExtractorConfig(FeatureExtractorConfig):
    _fit_process_desc: str = field(
        default="Determining the principal components", init=False, repr=False
    )
    _transform_process_desc: str = field(
        default="Transforming the input data to principal components",
        init=False,
        repr=False,
    )
    processor_name: str = field(default="pca", init=False, repr=False)

    input_columns: List[str] = None
    n_components: int = None
    copy: bool = True
    whiten: bool = False
    svd_solver: str = "auto"
    tol: float = 0.0
    iterated_power: str = "auto"
    n_oversamples: int = 10
    power_iteration_normalizer: str = "auto"
    random_state: int = None


class PCAFeatureExtractor(FeatureExtractor):
    output_dtype = "float64"
    config_class = PCAFeatureExtractorConfig
    config: PCAFeatureExtractorConfig

    def __init__(
        self,
        input_columns: List[str] = None,
        n_components: int = None,
        copy: bool = True,
        whiten: bool = False,
        svd_solver: str = "auto",
        tol: float = 0.0,
        iterated_power: str = "auto",
        n_oversamples: int = 10,
        power_iteration_normalizer: str = "auto",
        random_state: int = None,
        config: PCAFeatureExtractorConfig = None,
        **kwargs,
    ):
        super().__init__(
            config=config,
            input_columns=input_columns,
            n_components=n_components,
            copy=copy,
            whiten=whiten,
            svd_solver=svd_solver,
            tol=tol,
            iterated_power=iterated_power,
            n_oversamples=n_oversamples,
            power_iteration_normalizer=power_iteration_normalizer,
            random_state=random_state,
            **kwargs,
        )
        self.pca = PCA(
            n_components=self.config.n_components,
            copy=self.config.copy,
            whiten=self.config.whiten,
            svd_solver=self.config.svd_solver,
            tol=self.config.tol,
            iterated_power=self.config.iterated_power,
            n_oversamples=self.config.n_oversamples,
            power_iteration_normalizer=self.config.power_iteration_normalizer,
            random_state=self.config.random_state,
        )

    @sync_backup_config
    def set_params(self, **kwargs):
        self.config = self.config.replace_defaults(**kwargs)
        self.pca = PCA(
            n_components=self.config.n_components,
            copy=self.config.copy,
            whiten=self.config.whiten,
            svd_solver=self.config.svd_solver,
            tol=self.config.tol,
            iterated_power=self.config.iterated_power,
            n_oversamples=self.config.n_oversamples,
            power_iteration_normalizer=self.config.power_iteration_normalizer,
            random_state=self.config.random_state,
        )
        return self

    def fit(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        num_proc: int = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ) -> "PCAFeatureExtractor":
        self.config._input_columns = self._set_input_columns_and_arity(
            input_columns or self.config.input_columns
        )
        return self._process_fit(
            X,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        )

    def transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = None,
        raise_if_missing: bool = None,
        cache_output: bool = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        load_from_cache_file: bool = None,
        batched: bool = None,
        batch_size: int = None,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = None,
        num_proc: int = None,
        fingerprint: str = None,
    ):
        self._input_columns = self._set_input_columns_and_arity(
            input_columns or self.config.input_columns
        )
        return self._process_transform(
            X,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def fit_transform(
        self,
        X,
        input_columns: SelectedColumnTypes = None,
        keep_unused_columns: bool = True,
        raise_if_missing: bool = False,
        cache_output: bool = None,
        load_from_cache_file: bool = None,
        batched: bool = True,
        batch_size: int = 1000,
        batch_format: str = None,
        output_format: str = None,
        map_kwargs: dict = {"fn_kwargs": {}},
        num_proc: int = None,
        cache_dir: str = None,
        cache_file_name: str = None,
        fingerprint: str = None,
    ):
        return self.fit(
            X,
            input_columns=input_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            num_proc=num_proc,
            map_kwargs=map_kwargs,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            fingerprint=fingerprint,
        ).transform(
            X,
            input_columns=input_columns,
            keep_unused_columns=keep_unused_columns,
            raise_if_missing=raise_if_missing,
            cache_output=cache_output,
            cache_dir=cache_dir,
            cache_file_name=cache_file_name,
            load_from_cache_file=load_from_cache_file,
            batched=batched,
            batch_size=batch_size,
            batch_format=batch_format,
            output_format=output_format,
            map_kwargs=map_kwargs,
            num_proc=num_proc,
            fingerprint=fingerprint,
        )

    def _fit_sklearn(self, X: Union[pd.DataFrame, "pl.DataFrame", np.ndarray]):
        self.config.estimator = self.pca.fit(X)
        return self

    def _process_fit_output(self, input, out):
        self.config._n_features_out = self.config.estimator.n_components_
        return super()._process_fit_output(input, out)

    def _transform_sklearn(self, X: Union[pd.DataFrame, "pl.DataFrame", np.ndarray]):
        return self.config.estimator.transform(X)



import importlib
import re
import warnings
from collections import OrderedDict
from typing import List, Union

from biocore.utils.import_util import (
    is_biosets_available,
    is_transformers_available,
)
from biocore.utils.inspect import get_kwargs

from biofit.processing import ProcessorConfig
from biofit.visualization.plotting import PlotterConfig

PROCESSOR_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStat"),
        ("row_sum", "RowSumStat"),
        ("correlation", "CorrelationStat"),
        ("distance", "DistanceStat"),
        ("row_missingness", "RowMissingnessStat"),
        ("row_mean", "RowMeanStat"),
        ("col_missingness", "ColumnMissingnessStat"),
        ("col_mean", "ColumnMeanStat"),
        ("lightgbm", "LightGBMModel"),
        ("lasso", "LassoModel"),
        ("random_forest", "RandomForestModel"),
        ("logistic_regression", "LogisticRegressionModel"),
        ("pcoa", "PCoAFeatureExtractor"),
        ("pca", "PCAFeatureExtractor"),
        ("label_binarizing", "LabelBinarizer"),
        ("label_encoding", "LabelEncoder"),
        ("upsampling", "UpSampler"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelector"),
        ("log", "LogTransformer"),
        ("relative_abundance", "RelativeAbundanceScaler"),
        ("tmm", "TMMScaler"),
        ("css", "CumulativeSumScaler"),
        ("missing_labels", "MissingLabelsSampleFilter"),
        ("min_prevalence_sample_filter", "MinPrevalenceSampleFilter"),
        ("row_abundance", "AbundanceSampleFilter"),
    ]
)

PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotter"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorPlotter"),
        ("relative_abundance", "RelativeAbundancePlotter"),
        ("css", "CumulativeSumScalerPlotter"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotter"),
        ("row_abundance", "AbundanceSampleFilterPlotter"),
    ]
)

CONFIG_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CumulativeSumScalerConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

PLOTTER_CONFIG_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfig"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfig"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfig"),
    ]
)

PROCESSOR_CATEGORY_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "stat"),
        ("row_sum", "stat"),
        ("correlation", "stat"),
        ("distance", "stat"),
        ("row_missingness", "stat"),
        ("row_mean", "stat"),
        ("col_missingness", "stat"),
        ("col_mean", "stat"),
        ("lightgbm", "models"),
        ("lasso", "models"),
        ("random_forest", "models"),
        ("logistic_regression", "models"),
        ("pcoa", "preprocessing"),
        ("pca", "preprocessing"),
        ("label_binarizing", "preprocessing"),
        ("label_encoding", "preprocessing"),
        ("upsampling", "preprocessing"),
        ("min_prevalence_feature_selector", "preprocessing"),
        ("log", "preprocessing"),
        ("relative_abundance", "preprocessing"),
        ("tmm", "preprocessing"),
        ("css", "preprocessing"),
        ("missing_labels", "preprocessing"),
        ("min_prevalence_sample_filter", "preprocessing"),
        ("row_abundance", "preprocessing"),
    ]
)

PROCESSOR_TYPE_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "feature_extraction"),
        ("pca", "feature_extraction"),
        ("label_binarizing", "encoding"),
        ("label_encoding", "encoding"),
        ("upsampling", "resampling"),
        ("min_prevalence_feature_selector", "feature_selection"),
        ("log", "transformation"),
        ("relative_abundance", "scaling"),
        ("tmm", "scaling"),
        ("css", "scaling"),
        ("missing_labels", "filtering"),
        ("min_prevalence_sample_filter", "filtering"),
        ("row_abundance", "filtering"),
    ]
)

METABOLOMICS_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

KMER_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

MS1_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

MALDI_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfigForMaldi"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

BIODATA_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

RNA_SEQ_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

PROTEOMICS_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

METAGENOMICS_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfigForMetagenomics"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfigForMetagenomics"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfigForMetagenomics"),
        (
            "min_prevalence_feature_selector",
            "MinPrevalenceFeatureSelectorConfigForMetagenomics",
        ),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfigForMetagenomics"),
        ("css", "CumulativeSumScalerConfigForMetagenomics"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

MS2_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

OTU_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfigForOTU"),
        ("row_sum", "RowSumStatConfigForOTU"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfigForOTU"),
        ("row_missingness", "RowMissingnessStatConfigForOTU"),
        ("row_mean", "RowMeanStatConfigForOTU"),
        ("col_missingness", "ColumnMissingnessStatConfigForOTU"),
        ("col_mean", "ColumnMeanStatConfigForOTU"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfigForOTU"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfigForOTU"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfigForOTU"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfigForOTU"),
        ("log", "LogTransformerConfigForOTU"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfigForOTU"),
        ("css", "CumulativeSumScalerConfigForOTU"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfigForOTU"),
        ("row_abundance", "AbundanceSampleFilterConfigForOTU"),
    ]
)

GENOMICS_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfig"),
        ("row_missingness", "RowMissingnessStatConfig"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfig"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfig"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfig"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfig"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfig"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

SNP_MAPPING_NAMES = OrderedDict(
    [
        ("col_sum", "ColumnSumStatConfig"),
        ("row_sum", "RowSumStatConfig"),
        ("correlation", "CorrelationStatConfig"),
        ("distance", "DistanceStatConfigForSNP"),
        ("row_missingness", "RowMissingnessStatConfigForSNP"),
        ("row_mean", "RowMeanStatConfig"),
        ("col_missingness", "ColumnMissingnessStatConfigForSNP"),
        ("col_mean", "ColumnMeanStatConfig"),
        ("lightgbm", "LightGBMConfig"),
        ("lasso", "LassoConfig"),
        ("random_forest", "RandomForestConfig"),
        ("logistic_regression", "LogisticRegressionConfig"),
        ("pcoa", "PCoAFeatureExtractorConfig"),
        ("pca", "PCAFeatureExtractorConfig"),
        ("label_binarizing", "LabelBinarizerConfig"),
        ("label_encoding", "LabelEncoderConfig"),
        ("upsampling", "UpSamplerConfigForSNP"),
        ("min_prevalence_feature_selector", "MinPrevalenceFeatureSelectorConfigForSNP"),
        ("log", "LogTransformerConfig"),
        ("relative_abundance", "RelativeAbundanceScalerConfig"),
        ("tmm", "TMMScalerConfigForSNP"),
        ("css", "CLRScalerConfig"),
        ("imputation", "ImputationConfig"),
        ("missing_labels", "MissingLabelsSampleFilterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowSampleFilterConfigForSNP"),
        ("row_abundance", "AbundanceSampleFilterConfig"),
    ]
)

METABOLOMICS_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfig"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfig"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfig"),
    ]
)

KMER_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForGenomics"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfigForGenomics"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForGenomics"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForGenomics"),
    ]
)

MS1_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForProteomics"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForProteomics"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForProteomics"),
    ]
)

MALDI_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForMaldi"),
        ("relative_abundance", "RelativeAbundancePlotterConfigForMaldi"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForMaldi"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForProteomics"),
    ]
)

BIODATA_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfig"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfig"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfig"),
    ]
)

RNA_SEQ_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfig"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfig"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfig"),
    ]
)

PROTEOMICS_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForProteomics"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForProteomics"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForProteomics"),
    ]
)

METAGENOMICS_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        (
            "min_prevalence_feature_selector",
            "MinPrevalencePlotterConfigForMetagenomics",
        ),
        ("relative_abundance", "RelativeAbundancePlotterConfigForMetagenomics"),
        ("css", "CumulativeSumScalerPlotterConfigForMetagenomics"),
        (
            "min_prevalence_sample_filter",
            "MinPrevalenceRowPlotterConfigForMetagenomics",
        ),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForMetagenomics"),
    ]
)

MS2_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForProteomics"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfig"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForProteomics"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForProteomics"),
    ]
)

OTU_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForOTU"),
        ("relative_abundance", "RelativeAbundancePlotterConfigForOTU"),
        ("css", "CumulativeSumScalerPlotterConfigForOTU"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForOTU"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForOTU"),
    ]
)

GENOMICS_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForGenomics"),
        ("relative_abundance", "RelativeAbundancePlotterConfig"),
        ("css", "CumulativeSumScalerPlotterConfigForGenomics"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForGenomics"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForGenomics"),
    ]
)

SNP_PLOTTER_MAPPING_NAMES = OrderedDict(
    [
        ("pcoa", "PCoAFeatureExtractorPlotterConfig"),
        ("min_prevalence_feature_selector", "MinPrevalencePlotterConfigForSNP"),
        ("relative_abundance", "RelativeAbundancePlotterConfigForSNP"),
        ("css", "CumulativeSumScalerPlotterConfigForSNP"),
        ("min_prevalence_sample_filter", "MinPrevalenceRowPlotterConfigForSNP"),
        ("row_abundance", "AbundanceSampleFilterPlotterConfigForSNP"),
    ]
)


def config_class_to_model_type(config):
    """Converts a config class name to the corresponding model type"""
    for key, cls in CONFIG_MAPPING_NAMES.items():
        if cls == config:
            return key
    for key, cls in CONFIG_MAPPING._extra_content.items():
        if cls.__name__ == config:
            return key
    return None


class _LazyConfigMapping(OrderedDict):
    """
    A dictionary that lazily load its values when they are requested.
    """

    def __init__(self, mapping):
        self._mapping = mapping
        self._extra_content = {}
        self._modules = {}

    def __getitem__(self, key: str):
        if key in self._extra_content:
            return self._extra_content[key]
        if key not in self._mapping:
            raise KeyError(key)
        value = self._mapping[key]
        module_name = key.replace("-", "_")
        processor_category = PROCESSOR_CATEGORY_MAPPING_NAMES.get(key, "models")
        processor_type = PROCESSOR_TYPE_MAPPING_NAMES.get(key, None)
        try:
            if module_name not in self._modules:
                package = (
                    f"biofit.{processor_category}"
                    if not processor_type
                    else f"biofit.{processor_category}.{processor_type}"
                )
                self._modules[module_name] = importlib.import_module(
                    f".{module_name}", package
                )
        except ImportError:
            if is_transformers_available() and processor_category == "models":
                from transformers.models.auto.configuration_auto import (
                    model_type_to_module_name,
                )

                module_name = model_type_to_module_name(key)
                if module_name not in self._modules:
                    self._modules[module_name] = importlib.import_module(
                        f".{module_name}", "transformers.models"
                    )
            else:
                raise
        if hasattr(self._modules[module_name], value):
            return getattr(self._modules[module_name], value)
        try:
            biofit_module = importlib.import_module("biofit")
            return getattr(biofit_module, value)
        except AttributeError:
            if is_transformers_available():
                transformers_module = importlib.import_module("transformers")
                return getattr(transformers_module, value)
            raise

    def keys(self):
        return list(self._mapping.keys()) + list(self._extra_content.keys())

    def values(self):
        return [self[k] for k in self._mapping.keys()] + list(
            self._extra_content.values()
        )

    def items(self):
        return [(k, self[k]) for k in self._mapping.keys()] + list(
            self._extra_content.items()
        )

    def __iter__(self):
        return iter(list(self._mapping.keys()) + list(self._extra_content.keys()))

    def __contains__(self, item):
        return item in self._mapping or item in self._extra_content

    def register(self, key, value, exist_ok=False):
        """
        Register a new configuration in this mapping.
        """
        if key in self._mapping.keys() and (not exist_ok):
            raise ValueError(
                f"'{key}' is already used by a Transformers config, pick another name."
            )
        self._extra_content[key] = value


CONFIG_MAPPING = _LazyConfigMapping(CONFIG_MAPPING_NAMES)
PLOTTER_CONFIG_MAPPING = _LazyConfigMapping(PLOTTER_CONFIG_MAPPING_NAMES)
METABOLOMICS_MAPPING = _LazyConfigMapping(METABOLOMICS_MAPPING_NAMES)
KMER_MAPPING = _LazyConfigMapping(KMER_MAPPING_NAMES)
MS1_MAPPING = _LazyConfigMapping(MS1_MAPPING_NAMES)
MALDI_MAPPING = _LazyConfigMapping(MALDI_MAPPING_NAMES)
BIODATA_MAPPING = _LazyConfigMapping(BIODATA_MAPPING_NAMES)
RNA_SEQ_MAPPING = _LazyConfigMapping(RNA_SEQ_MAPPING_NAMES)
PROTEOMICS_MAPPING = _LazyConfigMapping(PROTEOMICS_MAPPING_NAMES)
METAGENOMICS_MAPPING = _LazyConfigMapping(METAGENOMICS_MAPPING_NAMES)
MS2_MAPPING = _LazyConfigMapping(MS2_MAPPING_NAMES)
OTU_MAPPING = _LazyConfigMapping(OTU_MAPPING_NAMES)
GENOMICS_MAPPING = _LazyConfigMapping(GENOMICS_MAPPING_NAMES)
SNP_MAPPING = _LazyConfigMapping(SNP_MAPPING_NAMES)
METABOLOMICS_PLOTTER_MAPPING = _LazyConfigMapping(METABOLOMICS_PLOTTER_MAPPING_NAMES)
KMER_PLOTTER_MAPPING = _LazyConfigMapping(KMER_PLOTTER_MAPPING_NAMES)
MS1_PLOTTER_MAPPING = _LazyConfigMapping(MS1_PLOTTER_MAPPING_NAMES)
MALDI_PLOTTER_MAPPING = _LazyConfigMapping(MALDI_PLOTTER_MAPPING_NAMES)
BIODATA_PLOTTER_MAPPING = _LazyConfigMapping(BIODATA_PLOTTER_MAPPING_NAMES)
RNA_SEQ_PLOTTER_MAPPING = _LazyConfigMapping(RNA_SEQ_PLOTTER_MAPPING_NAMES)
PROTEOMICS_PLOTTER_MAPPING = _LazyConfigMapping(PROTEOMICS_PLOTTER_MAPPING_NAMES)
METAGENOMICS_PLOTTER_MAPPING = _LazyConfigMapping(METAGENOMICS_PLOTTER_MAPPING_NAMES)
MS2_PLOTTER_MAPPING = _LazyConfigMapping(MS2_PLOTTER_MAPPING_NAMES)
OTU_PLOTTER_MAPPING = _LazyConfigMapping(OTU_PLOTTER_MAPPING_NAMES)
GENOMICS_PLOTTER_MAPPING = _LazyConfigMapping(GENOMICS_PLOTTER_MAPPING_NAMES)
SNP_PLOTTER_MAPPING = _LazyConfigMapping(SNP_PLOTTER_MAPPING_NAMES)


DATASET_TO_MAPPER = {
    "metabolomics": METABOLOMICS_MAPPING,
    "kmer": KMER_MAPPING,
    "ms1": MS1_MAPPING,
    "maldi": MALDI_MAPPING,
    "biodata": BIODATA_MAPPING,
    "rna-seq": RNA_SEQ_MAPPING,
    "proteomics": PROTEOMICS_MAPPING,
    "metagenomics": METAGENOMICS_MAPPING,
    "ms2": MS2_MAPPING,
    "otu": OTU_MAPPING,
    "genomics": GENOMICS_MAPPING,
    "snp": SNP_MAPPING,
}

DATASET_TO_MAPPER_NAMES = {
    "metabolomics": METABOLOMICS_MAPPING_NAMES,
    "kmer": KMER_MAPPING_NAMES,
    "ms1": MS1_MAPPING_NAMES,
    "maldi": MALDI_MAPPING_NAMES,
    "biodata": BIODATA_MAPPING_NAMES,
    "rna-seq": RNA_SEQ_MAPPING_NAMES,
    "proteomics": PROTEOMICS_MAPPING_NAMES,
    "metagenomics": METAGENOMICS_MAPPING_NAMES,
    "ms2": MS2_MAPPING_NAMES,
    "otu": OTU_MAPPING_NAMES,
    "genomics": GENOMICS_MAPPING_NAMES,
    "snp": SNP_MAPPING_NAMES,
}

DATASET_PLT_TO_MAPPER = {
    "metabolomics": METABOLOMICS_PLOTTER_MAPPING,
    "kmer": KMER_PLOTTER_MAPPING,
    "ms1": MS1_PLOTTER_MAPPING,
    "maldi": MALDI_PLOTTER_MAPPING,
    "biodata": BIODATA_PLOTTER_MAPPING,
    "rna-seq": RNA_SEQ_PLOTTER_MAPPING,
    "proteomics": PROTEOMICS_PLOTTER_MAPPING,
    "metagenomics": METAGENOMICS_PLOTTER_MAPPING,
    "ms2": MS2_PLOTTER_MAPPING,
    "otu": OTU_PLOTTER_MAPPING,
    "genomics": GENOMICS_PLOTTER_MAPPING,
    "snp": SNP_PLOTTER_MAPPING,
}

DATASET_PLT_TO_MAPPER_NAMES = {
    "metabolomics": METABOLOMICS_PLOTTER_MAPPING_NAMES,
    "kmer": KMER_PLOTTER_MAPPING_NAMES,
    "ms1": MS1_PLOTTER_MAPPING_NAMES,
    "maldi": MALDI_PLOTTER_MAPPING_NAMES,
    "biodata": BIODATA_PLOTTER_MAPPING_NAMES,
    "rna-seq": RNA_SEQ_PLOTTER_MAPPING_NAMES,
    "proteomics": PROTEOMICS_PLOTTER_MAPPING_NAMES,
    "metagenomics": METAGENOMICS_PLOTTER_MAPPING_NAMES,
    "ms2": MS2_PLOTTER_MAPPING_NAMES,
    "otu": OTU_PLOTTER_MAPPING_NAMES,
    "genomics": GENOMICS_PLOTTER_MAPPING_NAMES,
    "snp": SNP_PLOTTER_MAPPING_NAMES,
}


DATASET_PREPROCESSOR_MAPPING_NAMES = OrderedDict(
    [
        (
            "otu",
            [
                "row_abundance",
                "min_prevalence_feature_selector",
                "css",
            ],
        ),
        (
            "asv",
            [
                "row_abundance",
                "min_prevalence_feature_selector",
                "css",
            ],
        ),
        (
            "abundance",
            [
                "row_abundance",
                "min_prevalence_feature_selector",
                "css",
            ],
        ),
        (
            "metagenomics",
            [
                "row_abundance",
                "min_prevalence_feature_selector",
                "css",
            ],
        ),
        (
            "snp",
            [
                "min_prevalence_sample_filter",
                "min_prevalence_feature_selector",
            ],
        ),
        (
            "genomics",
            [
                "min_prevalence_sample_filter",
                "min_prevalence_feature_selector",
            ],
        ),
        (
            "maldi",
            [
                "min_prevalence_sample_filter",
                "relative_abundance",
                "min_prevalence_feature_selector",
            ],
        ),
        (
            "metabolomics",
            [
                "min_prevalence_sample_filter",
                "min_prevalence_feature_selector",
            ],
        ),
        (
            "proteomics",
            [
                "min_prevalence_sample_filter",
                "min_prevalence_feature_selector",
            ],
        ),
        (
            "transcriptomics",
            [
                "min_prevalence_sample_filter",
                "min_prevalence_feature_selector",
            ],
        ),
    ]
)


class _LazyLoadAllMappings(OrderedDict):
    """
    A mapping that will load all pairs of key values at the first access (either by indexing, requestions keys, values,
    etc.)

    Args:
        mapping: The mapping to load.
    """

    def __init__(self, mapping):
        self._mapping = mapping
        self._initialized = False
        self._data = {}

    def _initialize(self):
        if self._initialized:
            return
        warnings.warn(
            "ALL_PRETRAINED_CONFIG_ARCHIVE_MAP is deprecated and will be removed in v5 of Transformers. It does not contain all available model checkpoints, far from it. Checkout hf.co/models for that.",
            FutureWarning,
        )
        for processor_name, map_name in self._mapping.items():
            processor_category = PROCESSOR_CATEGORY_MAPPING_NAMES.get(
                processor_name, "models"
            )
            processor_type = PROCESSOR_TYPE_MAPPING_NAMES.get(processor_name, None)
            try:
                module_name = processor_name.replace("-", "_")
                package = (
                    f"biofit.{processor_category}"
                    if not processor_type
                    else f"biofit.{processor_category}.{processor_type}"
                )
                module = importlib.import_module(f".{module_name}", package)
                mapping = getattr(module, map_name)
            except ImportError:
                if is_transformers_available() and processor_category == "models":
                    from transformers.models.auto.configuration_auto import (
                        model_type_to_module_name,
                    )

                    module_name = model_type_to_module_name(processor_name)
                    module = importlib.import_module(
                        f".{module_name}", "transformers.models"
                    )
                    mapping = getattr(module, map_name)
            self._data.update(mapping)
        self._initialized = True

    def __getitem__(self, key):
        self._initialize()
        return self._data[key]

    def keys(self):
        self._initialize()
        return self._data.keys()

    def values(self):
        self._initialize()
        return self._data.values()

    def items(self):
        self._initialize()
        return self._data.keys()

    def __iter__(self):
        self._initialize()
        return iter(self._data)

    def __contains__(self, item):
        self._initialize()
        return item in self._data


def _get_class_name(model_class: Union[str, List[str]]):
    if isinstance(model_class, (list, tuple)):
        return " or ".join([f"[`{c}`]" for c in model_class if c is not None])
    return f"[`{model_class}`]"


def _list_model_options(indent, config_to_class=None, use_model_types=True):
    if config_to_class is None and (not use_model_types):
        raise ValueError(
            "Using `use_model_types=False` requires a `config_to_class` dictionary."
        )
    if use_model_types:
        if config_to_class is None:
            model_type_to_name = {
                model_type: f"[`{config}`]"
                for model_type, config in CONFIG_MAPPING_NAMES.items()
            }
        else:
            model_type_to_name = {
                model_type: _get_class_name(model_class)
                for model_type, model_class in config_to_class.items()
                if model_type in PROCESSOR_MAPPING_NAMES
            }
        lines = [
            f"{indent}- **{model_type}** -- {model_type_to_name[model_type]} ({PROCESSOR_MAPPING_NAMES[model_type]} model)"
            for model_type in sorted(model_type_to_name.keys())
        ]
    else:
        config_to_name = {
            CONFIG_MAPPING_NAMES[config]: _get_class_name(clas)
            for config, clas in config_to_class.items()
            if config in CONFIG_MAPPING_NAMES
        }
        config_to_model_name = {
            config: PROCESSOR_MAPPING_NAMES[model_type]
            for model_type, config in CONFIG_MAPPING_NAMES.items()
        }
        lines = [
            f"{indent}- [`{config_name}`] configuration class: {config_to_name[config_name]} ({config_to_model_name[config_name]} model)"
            for config_name in sorted(config_to_name.keys())
        ]
    return "\n".join(lines)


def replace_list_option_in_docstrings(config_to_class=None, use_model_types=True):
    def docstring_decorator(fn):
        docstrings = fn.__doc__
        if docstrings is None:
            return fn
        lines = docstrings.split("\n")
        i = 0
        while (
            i < len(lines) and re.search("^(\\s*)List options\\s*$", lines[i]) is None
        ):
            i += 1
        if i < len(lines):
            indent = re.search("^(\\s*)List options\\s*$", lines[i]).groups()[0]
            if use_model_types:
                indent = f"{indent}    "
            lines[i] = _list_model_options(
                indent, config_to_class=config_to_class, use_model_types=use_model_types
            )
            docstrings = "\n".join(lines)
        else:
            raise ValueError(
                f"The function {fn} should have an empty 'List options' in its docstring as placeholder, current docstring is:\n{docstrings}"
            )
        fn.__doc__ = docstrings
        return fn

    return docstring_decorator


class AutoConfig:
    """
    This is a generic configuration class that will be instantiated as one of the configuration classes of the library
    when created with the [`~AutoConfig.from_pretrained`] class method.

    This class cannot be instantiated directly using `__init__()` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoConfig is designed to be instantiated using the `AutoConfig.from_pretrained(pretrained_model_name_or_path)` method."
        )

    @classmethod
    def for_processor(
        cls, model_type: str, *args, experiment_type: str = None, **kwargs
    ):
        if is_biosets_available():
            from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
        else:
            EXPERIMENT_TYPE_ALIAS = {}

        experiment_type = EXPERIMENT_TYPE_ALIAS.get(experiment_type, experiment_type)
        if experiment_type is not None:
            if experiment_type in DATASET_TO_MAPPER:
                mapper = DATASET_TO_MAPPER[experiment_type]
                if model_type in mapper:
                    _config_class = mapper[model_type]
                    return _config_class(*args, **kwargs)
                raise ValueError(
                    f"Unrecognized model identifier: {model_type}. Should contain one of {', '.join(mapper.keys())}"
                )
        if model_type in CONFIG_MAPPING:
            _config_class = CONFIG_MAPPING[model_type]
            cls_kwargs = get_kwargs(kwargs, _config_class.__init__)
            return _config_class(*args, **cls_kwargs)
        elif (
            is_transformers_available()
            and PROCESSOR_CATEGORY_MAPPING_NAMES.get(model_type, "models") == "models"
        ):
            from transformers.models.auto.configuration_auto import (
                AutoConfig as HfAutoConfig,
            )

            return HfAutoConfig.for_model(model_type, *args, **kwargs)
        raise ValueError(
            f"Unrecognized model identifier: {model_type}. Should contain one of {', '.join(CONFIG_MAPPING.keys())}"
        )

    @classmethod
    @replace_list_option_in_docstrings()
    def from_pretrained(cls, pretrained_model_name_or_path, **kwargs):
        if is_transformers_available():
            from transformers.models.auto.configuration_auto import (
                AutoConfig as HfAutoConfig,
            )

            return HfAutoConfig.from_pretrained(pretrained_model_name_or_path, **kwargs)
        raise EnvironmentError(
            "Using `AutoConfig.from_pretrained` requires the transformers library to be installed. You can install it with `pip install transformers`."
        )

    @staticmethod
    def register(processor_type, config, exist_ok=False):
        """
        Register a new configuration for this class.

        Args:
            model_type (`str`): The model type like "bert" or "gpt".
            config ([`PretrainedConfig`]): The config to register.
        """
        types = ProcessorConfig
        if is_transformers_available():
            from transformers import PretrainedConfig

            types = (PretrainedConfig, ProcessorConfig)
        config_processor_type = getattr(
            config, "processor_type", getattr(config, "model_type", None)
        )
        if issubclass(config, types) and config_processor_type != processor_type:
            raise ValueError(
                f"The config you are passing has a `model_type` attribute that is not consistent with the model type you passed (config has {config_processor_type} and you passed {processor_type}. Fix one of those so they match!"
            )
        CONFIG_MAPPING.register(processor_type, config, exist_ok=exist_ok)


class AutoPlotterConfig:
    """
    This is a generic configuration class that will be instantiated as one of the configuration classes of the library
    when created with the [`~AutoPlotterConfig.from_pretrained`] class method.

    This class cannot be instantiated directly using `__init__()` (throws an error).
    """

    @classmethod
    def for_dataset(cls, dataset_name: str, *args, **kwargs):
        if is_biosets_available():
            from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
        else:
            EXPERIMENT_TYPE_ALIAS = {}

        dataset_name = EXPERIMENT_TYPE_ALIAS.get(dataset_name, dataset_name)
        if dataset_name in DATASET_PREPROCESSOR_MAPPING_NAMES:
            preprocessors = DATASET_PREPROCESSOR_MAPPING_NAMES[dataset_name]
            mapper = DATASET_PLT_TO_MAPPER[dataset_name]
            classes = []
            for preprocessor in preprocessors:
                _config_class = mapper[preprocessor]
                classes.append(_config_class(*args, **kwargs))
            return classes

        raise ValueError(
            f"Unrecognized dataset identifier: {dataset_name}. Should contain one of {', '.join(DATASET_PREPROCESSOR_MAPPING_NAMES.keys())}"
        )

    @classmethod
    def for_processor(
        cls, processor_type: str, *args, dataset_name: str = None, **kwargs
    ):
        if dataset_name is not None:
            if is_biosets_available():
                from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
            else:
                EXPERIMENT_TYPE_ALIAS = {}

            dataset_name = EXPERIMENT_TYPE_ALIAS.get(dataset_name, dataset_name)
            if dataset_name in DATASET_PLT_TO_MAPPER:
                mapper = DATASET_PLT_TO_MAPPER[dataset_name]
                if processor_type in mapper:
                    _config_class = mapper[processor_type]
                    return _config_class(*args, **kwargs)
                raise ValueError(
                    f"Unrecognized model identifier: {processor_type}. Should contain one of {', '.join(mapper.keys())}"
                )
        if processor_type in PLOTTER_CONFIG_MAPPING:
            _config_class = PLOTTER_CONFIG_MAPPING[processor_type]
            return _config_class(*args, **kwargs)
        elif (
            is_transformers_available()
            and PROCESSOR_CATEGORY_MAPPING_NAMES.get(processor_type, "models")
            == "models"
        ):
            from transformers.models.auto.configuration_auto import (
                AutoConfig as HfAutoConfig,
            )

            return HfAutoConfig.for_model(processor_type, *args, **kwargs)
        raise ValueError(
            f"Unrecognized model identifier: {processor_type}. Should contain one of {', '.join(PLOTTER_CONFIG_MAPPING.keys())}"
        )

    @staticmethod
    def register(processor_type, config, exist_ok=False):
        """
        Register a new configuration for this class.

        Args:
            model_type (`str`): The model type like "bert" or "gpt".
            config ([`PretrainedConfig`]): The config to register.
        """
        types = PlotterConfig
        config_processor_type = getattr(
            config, "processor_type", getattr(config, "model_type", None)
        )
        if issubclass(config, types) and config_processor_type != processor_type:
            raise ValueError(
                f"The config you are passing has a `model_type` attribute that is not consistent with the model type you passed (config has {config_processor_type} and you passed {processor_type}. Fix one of those so they match!"
            )
        PLOTTER_CONFIG_MAPPING.register(processor_type, config, exist_ok=exist_ok)


class AutoPreprocessorConfig(AutoConfig):
    """
    This is a generic configuration class that will be instantiated as one of the configuration classes of the library
    when created with the [`~AutoPreprocessorConfig.from_pretrained`] class method.

    This class cannot be instantiated directly using `__init__()` (throws an error).
    """

    @classmethod
    def for_dataset(cls, dataset_name: str, *args, **kwargs):
        if is_biosets_available():
            from biosets.packaged_modules import EXPERIMENT_TYPE_ALIAS
        else:
            EXPERIMENT_TYPE_ALIAS = {}

        dataset_name = EXPERIMENT_TYPE_ALIAS.get(dataset_name, dataset_name)
        if dataset_name in DATASET_PREPROCESSOR_MAPPING_NAMES:
            preprocessors = DATASET_PREPROCESSOR_MAPPING_NAMES[dataset_name]
            mapper = DATASET_TO_MAPPER[dataset_name]
            classes = []
            for preprocessor in preprocessors:
                _config_class = mapper[preprocessor]
                classes.append(_config_class(*args, **kwargs.get(preprocessor, {})))
            return classes

        raise ValueError(
            f"Unrecognized dataset identifier: {dataset_name}. Should contain one of {', '.join(DATASET_PREPROCESSOR_MAPPING_NAMES.keys())}"
        )



import copy
import inspect
from typing import Union

import numpy as np
import pandas as pd
from biocore import DataHandler
from sklearn.metrics import (
    confusion_matrix as sk_confusion_matrix,
)
from sklearn.metrics import (
    log_loss,
    multilabel_confusion_matrix,
)
from sklearn.metrics._classification import (
    _check_set_wise_labels,
    _check_zero_division,
    _nanaverage,
    _prf_divide,
)
from sklearn.utils.class_weight import compute_class_weight


def confusion_matrix(
    y_true, y_pred, labels=None, sample_weight=None, samplewise=False, normalize=None
):
    y_true = DataHandler.to_numpy(y_true)
    y_pred = DataHandler.to_numpy(y_pred)

    if y_true.ndim == 1 or y_true.shape[1] == 1:
        y_true = y_true.flatten()
    if y_pred.ndim == 1 or y_pred.shape[1] == 1:
        y_pred = (y_pred.flatten() > 0.5).astype(int)
    else:
        y_pred = np.argmax(y_pred, axis=1)

    if len(labels) < 3:
        mat = sk_confusion_matrix(
            y_true,
            y_pred,
            labels=labels
            if labels is not None and not isinstance(labels[0], str)
            else None,
            sample_weight=sample_weight,
            normalize=normalize,
        )

        df = pd.DataFrame(
            mat,
            columns=[f"Predicted {label}" for label in labels],
            index=[f"Actual {label}" for label in labels],
        )
        return df
    else:
        # create a len(labels) x len(labels) matrix
        mat = np.zeros((len(labels), len(labels)), dtype=int)
        for i, label in enumerate(labels):
            for j, pred in enumerate(labels):
                mat[i, j] = np.sum((y_true == i) & (y_pred == j))

        df = pd.DataFrame(
            mat,
            columns=[f"Predicted {label}" for label in labels],
            index=[f"Actual {label}" for label in labels],
        )
        return df


def specificity(
    y_true,
    y_pred,
    *,
    labels=None,
    pos_label=1,
    average=None,
    sample_weight=None,
    zero_division="warn",
):
    _check_zero_division(zero_division)
    labels = _check_set_wise_labels(y_true, y_pred, average, labels, pos_label)

    # Calculate tp_sum, pred_sum, true_sum ###
    samplewise = average == "samples"
    MCM = multilabel_confusion_matrix(
        y_true,
        y_pred,
        sample_weight=sample_weight,
        labels=labels,
        samplewise=samplewise,
    )
    tp_sum = MCM[:, 1, 1]
    tn_sum = MCM[:, 0, 0]
    fp_sum = MCM[:, 0, 1]
    fn_sum = MCM[:, 1, 0]
    true_sum = tp_sum + fn_sum
    false_sum = tn_sum + fp_sum

    if average == "micro":
        false_sum = np.array([false_sum.sum()])

    # Divide, and on zero-division, set scores and/or warn according to
    # zero_division:
    specificity = _prf_divide(
        tn_sum, false_sum, "specificity", "false", average, "specificity", zero_division
    )

    # Average the results
    if average == "weighted":
        weights = true_sum
    elif average == "samples":
        weights = sample_weight
    else:
        weights = None

    if average is not None:
        assert average != "binary" or len(specificity) == 1
        specificity = _nanaverage(specificity, weights=weights)
    if isinstance(specificity, (list, tuple, np.ndarray)) and len(specificity) > 1:
        specificity = specificity[-1]
    return specificity


def log_loss_weighted(y_true, y_pred, labels=None, class_weights=None):
    ytrue = DataHandler.to_numpy(y_true).flatten()
    if labels is None:
        labels = DataHandler.unique(DataHandler.to_numpy(ytrue).flatten())
    else:
        labels = DataHandler.to_numpy(labels).flatten()
        if set(labels) - set(ytrue):
            labels = DataHandler.unique(DataHandler.to_numpy(ytrue).flatten())

    if class_weights is None:
        class_weights = np.ones(len(labels))
    else:
        class_weights = compute_class_weight(class_weights, classes=labels, y=ytrue)

    ypred = DataHandler.to_numpy(y_pred)
    if ypred.ndim == 2 and ypred.shape[1] == 1:
        ypred = ypred.flatten()
    class_weights = dict(zip(labels, class_weights))
    sample_weights = np.array([class_weights[label] for label in ytrue])
    return log_loss(ytrue, ypred, sample_weight=sample_weights)


def calculate_metrics(
    metrics: Union[dict, callable], y_true, y_pred, sub_task, labels=None
):
    results = {}

    if labels is not None:
        labels = list(range(len(labels)))
    if callable(metrics):
        params = inspect.signature(metrics).parameters
        eval_kwargs = {}
        if "labels" in params:
            eval_kwargs["labels"] = labels

        results["custom_metric"].append(metrics(y_true, y_pred, **eval_kwargs))

    else:
        for metric_name, (metric_func, _) in metrics.items():
            if sub_task == "binary_classification":
                if metric_name in ["logloss", "logloss_weighted", "auc"]:
                    results[metric_name] = metric_func(
                        y_true, DataHandler.select_column(y_pred, 1)
                    )
                else:
                    results[metric_name] = metric_func(
                        y_true,
                        DataHandler.ge(DataHandler.select_column(y_pred, 1), 0.5),
                    )
            elif sub_task == "multiclass_classification":
                if metric_name in (
                    "accuracy",
                    "balanced_accuracy",
                    "f1_macro",
                    "f1_micro",
                    "f1_weighted",
                    "precision_macro",
                    "precision_micro",
                    "precision_weighted",
                    "recall_macro",
                    "recall_micro",
                    "recall_weighted",
                    "specificity_macro",
                    "specificity_micro",
                    "specificity_weighted",
                ):
                    results[metric_name] = metric_func(
                        y_true, DataHandler.argmax(y_pred, axis=1), labels
                    )
                else:
                    results[metric_name] = metric_func(y_true, y_pred, labels)
            else:
                if metric_name == "rmsle":
                    temp_pred = copy.deepcopy(DataHandler.to_numpy(y_pred))
                    temp_pred = np.clip(temp_pred, 0, None)
                    results[metric_name] = metric_func(y_true, temp_pred)
                else:
                    results[metric_name] = metric_func(y_true, y_pred)
    return results
